/*
	Purpose:	This module contains the main execution loop. The memory is
				organized like this:

					+-----------------+
					|		TEXT			|	128K
					|-----------------|
					|		DATA			|	128K
					|-----------------|
					|	OPERAND STACK	|
					|			...		|
					|	INTERNAL STACK	|	256K
					+-----------------+
	
				At this time heap allocation is not supported. Each instruction
				occupies one word (32 bits) and has the following format:
				
					31..8		7..0
					<data> <opcode>
	
	Author:	Alex Stojan
*/

/*
	TO-DO
	-----
	*	Revisit the UWCMPEQ/UWCMPNEQ concept and implementation
	*	Revisit the general implementation (using reinterpret_cast)
*/

#include <iostream>
#include <fstream>
#include <stack>
#include <memory.h>

#include "Globals.h"
#include "Loader.h"
#include "Utilities.h"
#include "AsmFileParser.h"
#include "SysCall.h"

using namespace std;

const INT32 MEM_SIZE	= 512000;

// This expression returns an integer number that represents the higher 24
// bits of an instruction.
//
#define VALUE_FIELD ((*reinterpret_cast<INT32*>(pc)) >> 8)

#define BASE (reinterpret_cast<INT32>(baseData))

CHAR8*	memPtr;

// base addresses
CHAR8*	baseData;
CHAR8*	baseText;
INT32*	baseStack;		// operand stack is 32-bit wide

// memory pointers to data and instructions
CHAR8*	dataPtr;
CHAR8*	textPtr;

// expression stack pointer
INT32*	x;

// stack (local area) pointer
CHAR8*	sp;

// stack for saving XSPs
INT32*	xspstack;

#ifdef _DEBUG
	INT32* x_start;
	CHAR8* sp_start;
	INT32* xsp_start;
#endif

//*******************************************************************

VM_STATUS_CODE init() {
	memPtr = static_cast<CHAR8*>(malloc(MEM_SIZE));

	if (memPtr) {
		baseText = textPtr = memPtr + 4;
		baseData = dataPtr = memPtr + 128000;			// 128 Kb
		baseStack = x = reinterpret_cast<INT32*>(memPtr + 256000);	// 256 Kb
		sp = reinterpret_cast<CHAR8*>(x);
		xspstack = reinterpret_cast<INT32*>(memPtr + 512000);

		dataPtr += WORD_SIZE;
	}
	else {
		return OUT_OF_MEMORY;
	}
	
#ifdef _DEBUG
	x_start = x;
	sp_start = sp;
	xsp_start = xspstack;
#endif
	
	return OK;
}

//*******************************************************************

void unload() {
	free(memPtr);
}

//*******************************************************************

void executionLoop() {
	CHAR8* pc = baseText;
	
	while (pc < textPtr) {
		int opcode = *pc;
		
		switch (opcode) {

			// Do nothing.
			//
			case IC_NOP: break;

			// ##### load/store instructions #####

			// Load byte at the given address onto operand stack. The loaded byte
			// is extended to 32 bits.
			//
			// Format: BLOAD <A> (<A> is the byte address)
			// Stack: ... => ..., W
			//
			case IC_BLOAD: {
				CHAR8* a = baseData + VALUE_FIELD;
				*(x++) = *a;
				break;
			}

			// Same as BLOAD, but with SP as the base.
			//
			case IC_SPBLOAD: {
				CHAR8* a = sp + VALUE_FIELD;
				*(x++) = *a;
				break;
			}

			// Load byte at address BASE + OFFSET onto operand stack. The rest is
			// the same as BLOAD. The base is assumed to be aligned to 32 bits.
			//
			// Format: BLOADI
			// Stack: ..., base, offset => ..., W
			//
			case IC_BLOADI: {
				INT32 offset = *(--x);
				INT32 base = *(--x);
				CHAR8* a = baseData + base + offset;
				*(x++) = *a;
				break;
			}

			// Load byte at address BASE (zero offset) onto operand stack. The
			// rest is the same as BLOAD. The base is assumed to be aligned to 32
			// bits.
			//
			// Format: BLOADIZ
			// Stack: ..., base => ..., W
			//
			case IC_BLOADIZ: {
				INT32 base = *(--x);
				CHAR8* a = baseData + base;
				*(x++) = *a;
				break;
			}

			// Store byte from the top of the operand stack at the given address.
			// The address is a byte-address, rather than word-address.
			//
			// Format: BSTORE <A> (<A> is the byte address)
			// Stack: ..., W => ...
			//
			case IC_BSTORE: {
				CHAR8* a = baseData + VALUE_FIELD;
				*a = *(--x);
				break;
			}

			// Same as BSTORE, but with SP as the base.
			//
			case IC_SPBSTORE: {
				CHAR8* a = sp + VALUE_FIELD;
				*a = *(--x);
				break;
			}

			// Store the byte value from the top of the operand stack at address
			// BASE + OFFSET.
			//
			// Format: BSTOREI
			// Stack: ..., base, offset, value => ...
			//
			case IC_BSTOREI: {
				CHAR8 value = *(--x);
				INT32 offset = *(--x);
				INT32 base = *(--x);
				CHAR8* a = baseData + base + offset;
				*a = value;
				break;
			}

			// Store the byte-value from the top of the operand stack at address
			// BASE (zero offset).
			//
			// Format: BSTOREIZ
			// Stack: ..., base, value => ...
			//
			case IC_BSTOREIZ: {
				CHAR8 value = *(--x);
				INT32 base = *(--x);
				CHAR8* a = baseData + base;
				*a = value;
				break;
			}

			// Load word from the given address onto operand stack.
			//
			// Format: WLOAD <A> (<A> is the word address)
			// Stack: ... => ..., W
			//
			case IC_WLOAD: {
				INT32* a = reinterpret_cast<INT32*>(baseData + VALUE_FIELD);
				*x++ = *a;
				break;
			}

			// Same as WLOAD, but with SP as the base.
			//
			case IC_SPWLOAD: {
				INT32* a = reinterpret_cast<INT32*>(sp + VALUE_FIELD);
				*x++ = *a;
				break;
			}

			// Load word at address BASE + OFFSET onto operand stack. The rest is
			// the same as LOAD.
			//
			// Format: WLOADI
			// Stack: ..., base, offset => ..., W
			//
			case IC_WLOADI: {
				INT32 offset = *(--x);
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base) + offset;
				*x++ = *a;
				break;
			}

			// Load word at address BASE (zero offset) onto operand stack. The
			// rest is the same as WLOAD.
			//
			// Format: WLOADIZ
			// Stack: ..., base => ..., W
			//
			case IC_WLOADIZ: {
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base);
				*x++ = *a;
				break;
			}

			// Store word from the top of the operand stack at the given address.
			//
			// Format: WSTORE <A> (<A> is the word address)
			// Stack: ..., W => ...
			//
			case IC_WSTORE: {
				INT32* a = reinterpret_cast<INT32*>(baseData + VALUE_FIELD);
				*a = *(--x);
				break;
			}

			// Same as WSTORE, but with SP as the base.
			//
			case IC_SPWSTORE: {
				INT32* a = reinterpret_cast<INT32*>(sp + VALUE_FIELD);
				*a = *(--x);
				break;
			}

			// Store the word from the top of the operand stack at address
			// BASE + OFFSET.
			//
			// Format: WSTOREI
			// Stack: ..., base, offset, value => ...
			//
			case IC_WSTOREI: {
				INT32 value = *(--x);
				INT32 offset = *(--x);
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base) + offset;
				*a = value;
				break;
			}

			// Store the word from the top of the operand stack at address BASE
			// (zero offset).
			//
			// Format: WSTOREIZ
			// Stack: ..., base, value => ...
			//
			case IC_WSTOREIZ: {
				INT32 value = *(--x);
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base);
				*a = value;
				break;
			}

			// Load the given integer constant onto operand stack.
			//
			// Format: ICLOAD <N>
			// Stack: ... => ..., W
			//
			case IC_ICLOAD: {
				*x++ = VALUE_FIELD;
				break;
			}

			// Increment an integer value at the given address by 1.
			//
			// Format: INC1 <A> (<A> is the word address)
			// Stack: ... => ...
			//
			case IC_INC1: {
				++*(reinterpret_cast<INT32*>(baseData + VALUE_FIELD));
				break;
			}
			
			// Same as INC1, but with SP as the base.
			//
			case IC_SPINC1: {
				++*(reinterpret_cast<INT32*>(sp + VALUE_FIELD));
				break;
			}

			// Increment an integer value at address BASE + OFFSET by VALUE.
			//
			// Format: INCI
			// Stack: ..., base, offset, value => ...
			//
			case IC_INCI: {
				INT32 value = *(--x);
				INT32 offset = *(--x);
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base) + offset;
				*a += value;
				break;
			}

			// Decrement an integer value at the given address by 1.
			//
			// Format: DEC1 <A> (<A> is the word address)
			// Stack: ... => ...
			//
			case IC_DEC1: {
				--*(reinterpret_cast<INT32*>(baseData + VALUE_FIELD));
				break;
			}
			
			// Same as DEC1, but with SP as the base.
			//
			case IC_SPDEC1: {
				--*(reinterpret_cast<INT32*>(sp + VALUE_FIELD));
				break;
			}

			// Decrement an integer value at address BASE + OFFSET by VALUE.
			//
			// Format: DECI
			// Stack: ..., base, offset, value => ...
			//
			case IC_DECI: {
				INT32 value = *(--x);
				INT32 offset = *(--x);
				INT32 base = *(--x);
				INT32* a = reinterpret_cast<INT32*>(baseData + base) + offset;
				*a -= value;
				break;
			}

			// ##### Stack manipulation instructions #####

			// Increment expression stack pointer by N words.
			//
			// Format: INCX <N>
			// Stack: ... => ...
			//
			case IC_INCX: {
				x += VALUE_FIELD;
				break;
			}

			// Decrement expression stack pointer by N words.
			//
			// Format: DECX <N>
			// Stack: ... => ...
			//
			case IC_DECX: {
				x -= VALUE_FIELD;
				break;
			}

			// Push X (the expression stack pointer, i.e. the operand stack 
			// pointer) onto the internal stack (this is not the expression 
			// stack).
			//
			// Format: PUSHX
			// Stack: ... => ...
			//
			case IC_PUSHX: {
				*xspstack-- = reinterpret_cast<INT32>(x);
				break;
			}
			
			// Pop X off the top of the internal stack and write it into SP.
			//
			// Format: POPXSP
			// Stack: ... => ...
			//
			case IC_POPXSP: {
				sp = reinterpret_cast<CHAR8*>(*++xspstack);
				break;
			}

			// Load SP onto the operand stack.
			//
			// Format: LOADSP
			// Stack: ... => ..., SP
			//
			case IC_LOADSP: {
				*x++ = reinterpret_cast<INT32>(sp - BASE);
				break;
			}

			// Read a value using SP as the base, add the base address of the
			// data area to it and write the value back into SP. This instruction
			// is used for restoring SP when a function returns.
			//
			// Format: SETSP <N>
			// Stack: ... => ...
			//
			case IC_SETSP: {
				INT32* a = reinterpret_cast<INT32*>(sp + VALUE_FIELD);
				sp = reinterpret_cast<CHAR8*>(*a + BASE);
				break;
			}

			// ##### Logic, arithmetic and relational instructions #####
								 
			// Logic, arithmetic and relational instructions can be binary and 
			// unary. Binary instructions take two operands from the top of the 
			// stack and push the result onto the stack. The unary instructions
			// take one operand from the top of the stack and push the result onto
			// the stack.
			//
			// Format: <instruction>
			// Stack (binary): ..., W, W => ..., W
			// Stack (unary): ..., W => ..., W
			//

			// ##### logic instructions #####

			case IC_AND:
			case IC_OR: {
				int top = *(--x);
				int next = *(--x);

				switch (opcode) {
					case IC_AND: *x++ = top && next;  break;
					case IC_OR: *x++ = top || next;  break;
				}

				break;
			}

			case IC_NOT: {
				int top = *(--x);
				*x++ = !top;
				break;
			}

			// ##### arithmetic instructions #####

			// Float addition, subtraction, multiplication, and division

			case IC_FADD:
			case IC_FSUB:
			case IC_FMUL:
			case IC_FDIV: {
				float* right = reinterpret_cast<float*>(--x);
				float* left = reinterpret_cast<float*>(--x);
				float* result = reinterpret_cast<float*>(x++);

				switch (opcode) {
					case IC_FADD: *result = *left + *right;  break;
					case IC_FSUB: *result = *left - *right;  break;
					case IC_FMUL: *result = *left * *right;  break;
					case IC_FDIV: *result = *left / *right;  break;
				}
				
				break;
			}

			// Float greater-or-equal, greater-than, less-or-equal, less-than
			
			case IC_FCMPGE:
			case IC_FCMPGT:
			case IC_FCMPLE:
			case IC_FCMPLT: {
				float* right = reinterpret_cast<float*>(--x);
				float* left = reinterpret_cast<float*>(--x);

				switch (opcode) {
					case IC_FCMPGE: *x++ = (*left >= *right ? 1 : 0);  break;
					case IC_FCMPGT: *x++ = (*left > *right ? 1 : 0);  break;
					case IC_FCMPLE: *x++ = (*left <= *right ? 1 : 0);  break;
					case IC_FCMPLT: *x++ = (*left < *right ? 1 : 0);  break;
				}
				
				break;
			}
			
			// Float negation

			case IC_FNEG: {
				float* spIn = reinterpret_cast<float*>(--x);
				float* spOut = reinterpret_cast<float*>(x++);
				*spOut = -*spIn;
				break;
			}
			
			// Integer addition, subtraction, multiplication, and division

			case IC_IADD:
			case IC_ISUB:
			case IC_IMUL:
			case IC_IDIV: {
				int right = *(--x);
				int left = *(--x);

				switch (opcode) {
					case IC_IADD: *x++ = left + right;  break;
					case IC_ISUB: *x++ = left - right;  break;
					case IC_IMUL: *x++ = left * right;  break;
					case IC_IDIV: *x++ = left / right;  break;
				}

				break;
			}

			// Integer greater-or-equal, greater-than, less-or-equal, less-than

			case IC_ICMPGE:
			case IC_ICMPGT:
			case IC_ICMPLE:
			case IC_ICMPLT: {
				int right = *(--x);
				int left = *(--x);

				switch (opcode) {
					case IC_ICMPGE: *x++ = (left >= right);  break;
					case IC_ICMPGT: *x++ = (left > right);  break;
					case IC_ICMPLE: *x++ = (left <= right);  break;
					case IC_ICMPLT: *x++ = (left < right);  break;
				}
				
				break;
			}

			// Word comparison (equal and not-equal), used for floats and integers
			//
			case IC_WCMPEQ:
			case IC_WCMPNEQ: {
				INT32 right = *(--x);
				INT32 left = *(--x);
				
				switch (opcode) {
					case IC_WCMPEQ: *x++ = (left == right);  break;
					case IC_WCMPNEQ: *x++ = (left != right);  break;
				}
				
				break;
			}
			
			// Integer negation

			case IC_INEG: {
				INT32* spIn = --x;
				*x++ = *spIn * -1;
				break;
			}
			
			// Modulo (remainder of division of two integers)

			case IC_MOD: {
				int right = *(--x);
				int left = *(--x);
				*x++ = left % right;
				break;
			}

			// ##### branch/jump instructions #####

			// All branch and jump instruction specify their offset in number of
			// words rather than bytes, so they can specify a larger distance to
			// jump to. The WORD_SIZE is subtracted because the PC will be 
			// incremented at the end of the execution loop.

			// Jump to location N using the base address of the text area.
			//
			// Format: GOTO <N>
			//
			case IC_GOTO: {
				INT32 target = VALUE_FIELD;
				pc = baseText + target * WORD_SIZE - WORD_SIZE;
				break;
			}

			// Jump to location stored at the top of the operand stack. The rest
			// is the same as for GOTO.
			//
			// Format: GOTOI
			// Stack: ..., loc => ...
			//
			case IC_GOTOI: {
				pc = baseText + *(--x) * WORD_SIZE - WORD_SIZE;
				break;
			}

			// Pop a value off the top of the operand stack and if it's a non-zero
			// value jump to location N (as done for GOTO).
			//
			// Format: IFTRUE <N>
			// Stack: ..., value => ...
			//
			case IC_IFTRUE: {
				if (*(--x)) {
					INT32 target = VALUE_FIELD;
					pc = baseText + target * WORD_SIZE - WORD_SIZE;
				}
				
				break;
			}

			// Similar to IFTRUE, but tests for a zero-value.
			//
			// Format: IFFALSE <N>
			// Stack: ..., value => ...
			//
			case IC_IFFALSE: {
				if (! *(--x)) {
					INT32 target = VALUE_FIELD;
					pc = baseText + target * WORD_SIZE - WORD_SIZE;
				}
				
				break;
			}

			// ##### conversion instructions #####

			// Convert float to integer. The float value is on the top of the
			// operand stack.
			//
			// Format: FTOI
			// Stack: ..., F => ..., I
			//
			case IC_FTOI: {
				float* sp = reinterpret_cast<float*>(--x);
				*x++ = static_cast<int>(*sp);
				break;
			}

			// Convert integer to float. The integer value is on the top of the
			// operand stack.
			//
			// Format: ITOF
			// Stack: ..., I => ..., F
			//
			case IC_ITOF: {
				int top = *(--x);
				float* sp = reinterpret_cast<float*>(x++);
				*sp = static_cast<float>(top);
				break;
			}

			// ##### system call instruction(s) #####
			
			// Format: SYSCALL <system code>, <arg count>
			// Stack: ..., value, type, value, type, value, type => ...
			//
			case IC_SYSCALL: {
				CHAR8 sysCode = *(pc + 1);
				CHAR8 argCount = *(pc + 2);

				stack<SysCallPair> argStack;

				// The I/O arguments appear on the stack in the reversed
				// order, but we need to process them in the original order.
				//
				for (int i = 0;  i < argCount;  ++i) {
					INT32 type = *(--x);
					INT32* value = --x;
					
					argStack.push(SysCallPair(type, value));
				}
				
				SYS_CALL_RESULT result = sysCall(sysCode, argStack, baseData);
				if (result != SUCCESS) {
					cout << "\n\n% ERROR: System call failure (" << sysCode << ")\n\n" ;
					exit(0);
				}

				break;
			}

			default:	{
				cout << "\n\n% ERROR: Undefined instruction code (" << opcode << ")\n\n" ;
				exit(0);
			}
		}

		pc += WORD_SIZE;
	}

#ifdef _DEBUG
	if (x != x_start) {
		cout << "\n!!! WARNING: X = " << x << " (initial value: " << x_start << ")";
	}
		
	if (sp != sp_start) {
		void* spBefore = sp_start;
		void* spAfter = sp;
		cout << "\n!!! WARNING: SP = " << static_cast<INT32*>(spAfter) << " (initial value: " << static_cast<INT32*>(spBefore) << ")";
	}
	
	if (xspstack != xsp_start) {
		cout << "\n!!! WARNING: XSP STACK = " << x << " (initial value: " << xsp_start << ")";
	}
#endif
}

//*******************************************************************

int main(int argc, char* argv[]) {
	ofstream* out = 0;
	
	if (argv[2] != 0) {
		out = new ofstream(argv[2]);
	}

	init();

	INT32* data = reinterpret_cast<INT32*>(dataPtr);
	INT32* text = reinterpret_cast<INT32*>(textPtr);
	
	loadAsmFile(string(argv[1]), data, text, out);

	dataPtr = reinterpret_cast<CHAR8*>(data);
	textPtr = reinterpret_cast<CHAR8*>(text);
	
	if (out) *out << "\n" << static_cast<int>(textPtr - baseText) << " bytes";

	executionLoop();
	delete out;

	return 0;
}