#include <string>
#include <math.h>
#include <iostream>
#include <vector>

#include "Globals.h"
#include "Utilities.h"
#include "Loader.h"

using namespace std;

//*******************************************************************
//*******************************************************************
//*******************************************************************

int encodeInstruction(int opcode, int i1) {
	return opcode | (i1 << 8);
}

//*******************************************************************

int encodeInstruction(int opcode, int i1, int i2) {
	switch (opcode) {
		case IC_SYSCALL:
			return opcode | (i1 << 8) | (i2 << 16);
		
		default:
			return opcode | (i1 << 8) | (i2 << 20);
	}
}

//*******************************************************************

void loadCharString(const char* s, INT32*& dataPtr, ofstream* out) {
	size_t len = strlen(s) + 1;		// we need space for 0, so we add 1
	float size = len / 4.0f;
	int wordCount = static_cast<int>(ceil(size));

	memcpy(dataPtr, s, len);

	if (out) {
		*out << "\n.asciiz \"" << s << "\" (" << wordCount * WORD_SIZE << ") @" << dataPtr;
		for (int i = 0;  i < wordCount;  ++i) {
			showWord(dataPtr + i, true);
		}
	}

	dataPtr += wordCount;
}

//*******************************************************************

void loadFloatValue(float v, INT32*& dataPtr, ofstream* out) {
	float* a = reinterpret_cast<float*>(dataPtr++);
	*a = v;

	if (out) {
		*out << "\n.float " << *a << " @" << a;
	}
}

//*******************************************************************

void allocateStaticBlock(int size, INT32*& dataPtr, ofstream* out) {
	int wordCount = static_cast<int>(ceil(size / 4.0));

	if (out) {
		*out << "\n.block " << wordCount * WORD_SIZE << " @" << dataPtr;
	}

	dataPtr += wordCount;
}

//*******************************************************************

void loadData(vector<DataItem> data, INT32*& dataPtr, ofstream* out) {
	for (size_t i = 0;  i < data.size();  ++i) {
		DataItem item = data[i];
		
		switch (item.type) {
			case DataItem::STRING:	loadCharString(item.str.c_str(), dataPtr, out);  break;
			case DataItem::FLOAT:	loadFloatValue(item.floatVal, dataPtr, out);  break;
			case DataItem::BLOCK:	allocateStaticBlock(item.block, dataPtr, out);  break;
		}
	}
}

//*******************************************************************

void loadCode(vector<Instruction> code, INT32*& textPtr, ofstream* out) {
	for (size_t i = 0;  i < code.size();  ++i) {
		Instruction instr = code[i];
		int formatted = 0;
		switch (instr.opcode) {
			// one-index instructions
			case IC_ICLOAD:
			case IC_INC1:
			case IC_SPINC1:
			case IC_DEC1:
			case IC_SPDEC1:
			case IC_WLOAD:
			case IC_BLOAD:
			case IC_WSTORE:
			case IC_BSTORE:
			case IC_SPBSTORE:
			case IC_GOTO:
			case IC_IFTRUE:
			case IC_IFFALSE:
			case IC_INCX:
			case IC_DECX:
			case IC_SPWLOAD:
			case IC_SPBLOAD:
			case IC_SPWSTORE:
			case IC_SETSP: {
				formatted = encodeInstruction(instr.opcode, instr.optr1);
				break;
			}
			
			// two-index instructions
			case IC_SYSCALL: {
				formatted = encodeInstruction(instr.opcode, instr.optr1, instr.optr2);
				break;
			}

			// simple (no index) instructions
			case IC_NOP:
			case IC_INCI:
			//case IC_INCISP:
			case IC_DECI:
			//case IC_DECISP:
			case IC_BSTOREI:
			case IC_BSTOREIZ:
			case IC_WSTOREI:
			case IC_WSTOREIZ:
			case IC_IADD:
			case IC_ISUB:
			case IC_IMUL:
			case IC_IDIV:
			case IC_MOD:
			case IC_OR:
			case IC_AND:
			case IC_ICMPLT:
			case IC_ICMPGT:
			case IC_ICMPLE:
			case IC_ICMPGE:
			case IC_INEG:
			case IC_NOT:
			case IC_WCMPEQ:
			case IC_ITOF:
			case IC_FADD:
			case IC_FSUB:
			case IC_FMUL:
			case IC_FDIV:
			case IC_FCMPLT:
			case IC_FCMPGT:
			case IC_FCMPLE:
			case IC_FCMPGE:
			case IC_FNEG:
			case IC_WCMPNEQ:
			case IC_WLOADI:
			//case IC_WLOADISP:
			case IC_WLOADIZ:
			//case IC_WLOADIZSP:
			case IC_BLOADI:
			case IC_BLOADIZ:
			//case IC_BLOADISP:
			//case IC_BLOADIZSP:
			//case IC_BSTOREISP:
			//case IC_BSTOREIZSP:
			//case IC_WSTOREISP:
			//case IC_WSTOREIZSP:
			case IC_FTOI:
			case IC_GOTOI: 
			case IC_PUSHX:
			case IC_POPXSP:
			case IC_LOADSP: {
				formatted = instr.opcode;
				break;
			}
		}

		INT32* addr = textPtr++;
		*addr = formatted;
		
		if (out) showInstrMem(addr, instr, out);
	}
}