#include <iostream>
#include <fstream>
#include <cmath>
#include <deque>
#include <algorithm>

#include "Symbols.h"

using namespace boost;
using namespace std;
using namespace Compiler;

//*******************************************************************
//*******************************************************************

//#################### Global Functions ####################

// Returns the next word address based on index and size. It is used to align
// an index to the next word when calculating data sizes.
//
int alignIndex(int index, int size) {
	return static_cast<int>(ceil((index + static_cast<double>(size)) / WORD_SIZE) * WORD_SIZE);
}

//*******************************************************************

//#################### TypeAttr ####################

//*******************************************************************

void TypeAttr::calcSize(string name) {
	if (isPtr) {
		size = WORD_SIZE;
	}
	else {
		size = getTypeSizeFromTypeName(name);

		if (isArray()) {
			for (size_t i = 0;  i < arrayDimensions.size();  ++i) {
				size *= arrayDimensions[i];
			}
		}
	}
}

//*******************************************************************

bool TypeAttr::setTypeName(std::string name, DimVector arrayDim) {
	typeName = name;
	arrayDimensions = arrayDim;
	calcSize(name);
	return setTypeNameCode();
}

//*******************************************************************

int TypeAttr::getSize(SizeMetric metric, TypeSizeSpec spec) const {
	if (spec == ARRAY_AS_POINTER && isArray()) {
		return getTypeSizeFromTypeName(INT, metric);
	}
	else {
		if (metric == WORD) {
			return static_cast<int>(ceil(static_cast<double>(size) / WORD_SIZE));
		}
		else {
			return size;
		}
	}
}

//*******************************************************************

bool TypeAttr::setTypeNameCode() {
	bool result = true;
	
	if (typeName == CHAR) {
		if (isOneDimensionalArray() || isPtr) {
			code = STRING_CODE;
		}
		else {
			code = CHAR_CODE;
		}
	}
	else if (typeName == INT) code = INT_CODE;
	else if (typeName == FLOAT) code = FLOAT_CODE;
	else if (typeName == BOOL) code = BOOL_CODE;
	else if (typeName == VOID) code = VOID_CODE;
	else
		result = false;	// undefined type
	
	return result;
}

//*******************************************************************

TypeCode TypeAttr::getStorageClass(bool ptrAsArray) const {
	if (isPtr && !ptrAsArray) {
		return INT_CODE;
	}
	else
		return code;
}

//*******************************************************************

void TypeAttr::reduceArray() {
	if (isPtr && !isArray()) {
		// after indexing a pointer, it bacomes a basic type
		isPtr = false;
		--refLevel;
	}
	else if (isArray() && arrayDimensions.size() > 0) {
		arrayDimensions.erase(arrayDimensions.begin());
	}
}

//*******************************************************************

bool TypeAttr::reduceRefLevel() {
	if (refLevel == 0) {
		return false;
	}
	
	--refLevel;
	if (refLevel == 0) {
		isPtr = false;
		
		// for example, char^ (which is of size 32 bits) turns into char (which
		// is of size 8 bits)
		size = getTypeSizeFromTypeName(typeName);
	}
	
	return true;
}

//*******************************************************************

void TypeAttr::increaseRefLevel() {
	++refLevel;
	isPtr = true;
	size = getTypeSizeFromTypeName(INT);	// pointer size
}

//*******************************************************************

//#################### Symbol ####################

//*******************************************************************

void Symbol::setParams(ParamVector p) {
	int size = 0;
	for (ParamVector::iterator itr = p.begin();  itr != p.end();  ++itr) {
		size += (*itr)->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER);
	}
	
	params = p;
	paramsSpace = size;
}

//*******************************************************************

//#################### SymbolTable ####################

//*******************************************************************

// static variables

SymbolTable::ConstMap SymbolTable::constTable;
int SymbolTable::ID = 0;

//*******************************************************************

SymbolTable::SymbolTable(): space(0), function(0) {
	id = ID++;
}

//*******************************************************************

SymbolTable::~SymbolTable() {
	for (SymbolPtrMap::iterator itr = fnTable.begin();  itr != fnTable.end();  ++itr) {
		delete itr->second;		// delete the Symbol object
	}
}

//*******************************************************************

bool SymbolTable::addFunction(Symbol* fn) {
	// check for duplicates
	for (SymbolSequence::iterator t = varSeq.begin();  t != varSeq.end(); ) {
		Symbol* sym = *t;

		if (*sym == *fn) {	// function with the same name found
			if (sym->isForwardedFunction) {
				if (! fn->isForwardedFunction) {
					// Verify that the forward declaration and the definition match
					// in types and parameters and, if they do, delete the forward
					// declaration.

					if (matchFwdToDef(*sym, *fn)) {
						varSeq.erase(t);
					}
					else
						return false;		// ERROR: non-matching parameter types
				}
				else
					return false;	// ERROR: duplicate forward declaration
			}
			else	// ERROR: duplicate function definition or forward declaration appearing after definition
				return false;
		}
		else
			++t;
	}

	fnTable[fn->name] = fn;
	varSeq.push_back(fn);
	
	if (! fn->isForwardedFunction) {
		space += fn->type.getSize(WORD);
	}
	
	return true;
}

//*******************************************************************

Symbol* SymbolTable::addSymbol(string name, TypeAttr type, bool isParameter) {
	SymbolMap::iterator itr = varTable.find(name);
	if (itr != varTable.end()) {
		return 0;	// a variable with this name already exists
	}

	varTable[name] = Symbol(name, id, type);
	Symbol* sym = &(varTable[name]);
	varSeq.push_back(sym);
	space += type.getSize(WORD, isParameter ? TypeAttr::ARRAY_AS_POINTER : TypeAttr::REGULAR);

	return sym;
}

//*******************************************************************

Symbol* SymbolTable::getSymbol(string name) {
	SymbolMap::iterator itr = varTable.find(name);
	
	if (itr != varTable.end()) {
		return &(itr->second);
	}
	
	return 0;
}

//*******************************************************************

Symbol* SymbolTable::getFunction(string name) {
	SymbolPtrMap::iterator itr = fnTable.find(name);
	
	if (itr != fnTable.end()) {
		return itr->second;
	}
	
	return 0;
}

//*******************************************************************

void SymbolTable::addConstant(Constant cns) {
	if (cns.type != TypeAttr(INT)) {
		constTable[cns.token] = cns;
	}
}

//*******************************************************************

//#################### Scope ####################

//*******************************************************************

SymbolTable* Scope::enter(Symbol* function) {
	SymbolTable st;
	allSymbolTables[st.getID()] = st;
	SymbolTable* top = &(allSymbolTables[st.getID()]);
	top->function = function;

	if (stStack.size() == 0) {
		topLevel = top;
	}

	stStack.push(top);
	
	return top;
}

//*******************************************************************

int Scope::exit() {
	space += stStack.top()->getSpace();
	stStack.pop();
	
	return space;
}

//*******************************************************************

//#################### Namespace Functions ####################

//*******************************************************************

void Compiler::generateConstantHeader(int* index, vector<string>* header, ofstream& debugFile) {
	for (SymbolTable::ConstMap::iterator it = SymbolTable::constTable.begin();
			it != SymbolTable::constTable.end();
			++it) {
		Constant& c = it->second;

		if (c.type.getName() == CHAR) {
			if (c.type.isArray()) {		// it's a string
				header->push_back(string(ASCIIZ_DEF) + " " + c.token);
			}

			c.index = *index;
			debugFile << c;
			*index = alignIndex(*index, c.type.getSize());
		}
		else if (c.type.getName() == FLOAT) {
			if (c.type.isArray()) {
				header->push_back(string(BLOCK_DEF) + " " + lexical_cast<string>(static_cast<int>(c.type.getSize())));
			}
			else {
				header->push_back(string(FLOAT_CONST) + " " + c.token);
			}

			c.index = *index;
			debugFile << c;
			*index = alignIndex(*index, c.type.getSize());
		}
	}
}

//*******************************************************************

// -> will NOT pack bytes together!
//
void Compiler::generateDataHeader(const Scope::SymbolTableMap& allSymbolTables, ofstream& asmFile, ofstream& debugFile) {
	vector<string> dataHeader;

	debugFile << "*** SYMBOLS ***\n\n";
	asmFile << DATA_SEGMENT << endl;
	int index = WORD_SIZE;
	
	generateConstantHeader(&index, &dataHeader, debugFile);

	for (Scope::SymbolTableMap::const_iterator itr = allSymbolTables.begin();
			itr != allSymbolTables.end();
			++itr) {

		const SymbolTable* st = &(itr->second);

		for (SymbolTable::SymbolSequence::const_iterator varItr = st->getVarSeq().begin();
				varItr != st->getVarSeq().end();
				++varItr) {

			Symbol* sym = *varItr;
			
			if (sym->isFunction) {
				debugFile << *sym;
				continue;
			}

			if (sym->type.getName() == BOOL || sym->type.getName() == CHAR) {
				if (sym->isLocal) {
					sym->index = st->function->localAddressCounter;
					debugFile << *sym;
					int size = (sym->type.isArray() && sym->isParameter) ? WORD_SIZE : sym->type.getSize();
					st->function->localAddressCounter = alignIndex(st->function->localAddressCounter, size);
				}
				else {
					if (sym->type.isArray() && ! sym->isParameter) {
						dataHeader.push_back(string(BLOCK_DEF) + " " + lexical_cast<string>(static_cast<int>(sym->type.getSize())));
					}
				
					sym->index = index;
					debugFile << *sym;
					index = alignIndex(index, sym->type.getSize());
				}
			}
			else if (sym->type.getName() == INT || sym->type.getName() == FLOAT) {
				if (sym->isLocal) {
					sym->index = st->function->localAddressCounter;
					debugFile << *sym;
					int size = (sym->type.isArray() && sym->isParameter) ? WORD_SIZE : sym->type.getSize();
					st->function->localAddressCounter = alignIndex(st->function->localAddressCounter, size);
				}
				else {
					if (sym->type.isArray() && ! sym->isParameter) {
						dataHeader.push_back(string(BLOCK_DEF) + " " + lexical_cast<string>(static_cast<int>(sym->type.getSize())));
					}
				
					sym->index = index;
					debugFile << *sym;
					index = alignIndex(index, sym->type.getSize());
				}
			}
		}
	}

	for (size_t i = 0;  i < dataHeader.size();  ++i) {
		asmFile << dataHeader[i] << endl;
	}
}

//*******************************************************************

Symbol* Compiler::findFunction(string name, const Scope& scope) {
	SymbolTable* st = scope.getTopLevel();
	return st->getFunction(name);
}

//*******************************************************************
	
Symbol* Compiler::findSymbol(string name, int stID, Scope::SymbolTableMap& allSymbolTables) {
	SymbolTable* st = &(allSymbolTables[stID]);
	return st->getSymbol(name);
}

//*******************************************************************

Symbol* Compiler::findSymbol(string name, const stack<SymbolTable*>& stStack) {
	stack<SymbolTable*>::container_type container = stStack.c;
	for (stack<SymbolTable*>::container_type::reverse_iterator itr = container.rbegin();
			itr != container.rend();
			++itr) {
		SymbolTable* st = *itr;
		Symbol* sym = st->getSymbol(name);
		if (sym) return sym;
	}

	return 0;	// not found
}

//*******************************************************************

// This function assumes the scanner has already parsed the constant correctly.
//
Constant Compiler::getNumericConstant(string c) {
	if (c.find(".") != string::npos) {
		return Constant(lexical_cast<float>(c), TypeAttr(FLOAT));
	}
	else {
		return Constant(lexical_cast<int>(c), TypeAttr(INT));
	}
}

//*******************************************************************

bool Compiler::matchFwdToDef(const Symbol& fwd, const Symbol& def) {
	if (fwd.fwdParamTypes.size() != def.params.size()) {
		return false;		// different number of parameters
	}
	else if (fwd.type != def.type) {
		return false;		// different return types
	}
	else {
		for (size_t i = 0;  i < def.params.size();  ++i) {
			Symbol* defParam = def.params[i];
			TypeAttr defParamType = defParam->type;
			TypeAttr fwdParamType = fwd.fwdParamTypes[i];
			
			if (defParamType != fwdParamType) {
				return false;		// different parameter types
			}
		}
	}
	
	return true;	// match!
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, const TypeAttr& attr) {
	string dims;
	const TypeAttr::DimVector& dim = attr.getArrayDim();
	for (size_t i = 0;  i < dim.size();  ++i) {
		dims += " " + lexical_cast<string>(dim[i]);
	}
	
	string str = "[CODE: " + attr.getName() + 
						", REFERENCE: " + lexical_cast<string>(attr.isPointer()) +
						", LEVEL: " + lexical_cast<string>(attr.getRefLevel()) +
						", ARRAY: " + lexical_cast<string>(attr.isArray()) +
						", DIM:" + dims +
						"]";

	return os << str;
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, const Symbol& sym) {
	return os << "\nNAME: '" << sym.name << "'   TYPE: " << 
				sym.type << "   INDEX: " << sym.index << 
				"   FUNCTION: " << sym.isFunction << "   LOCAL: " << sym.isLocal <<
				"   SPACE: " << sym.localSpace;
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, SymbolTable& st) {
	ostream& txt = os << "\n\nSYMBOL TABLE\n------------";

	for (SymbolTable::SymbolMap::const_iterator itr = st.getVarTable().begin();
			itr != st.getVarTable().end();
			++itr) {
		const Symbol* sym = &(itr->second);
		txt << "\nIDENTIFIER: '" << itr->first << "'  TYPE: " << sym->type << "  SIZE: " << sym->type.getSize();
	}

	for (SymbolTable::ConstMap::iterator itr = st.constTable.begin();
			itr != st.constTable.end();
			++itr) {
		Constant& c = itr->second;
		txt << "\nCONSTANT: " << c.token;
	}

	return txt;
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, const Constant& cns) {
	ostream& txt = os << "\nCONSTANT: " << cns.token << "   TYPE: " << cns.type << "   INDEX: " << cns.index << "   VALUE: ";

	if (cns.type.isArray()) {		// string constant
		return txt;
	}
	else {
		if (cns.type == TypeAttr(INT)) {
			return os << cns.intVal;
		}
		else if (cns.type == TypeAttr(FLOAT)) {
			return os << cns.floatVal;
		}
		else if (cns.type == TypeAttr(CHAR)) {
			return os << static_cast<char>(cns.intVal);
		}
		else if (cns.type == TypeAttr(BOOL)) {
			return os << cns.intVal;
		}
	}

	return txt;
}
