/*
	Purpose	:	This file contains all symbol-related functionality, i.e.
					symbol table, type, constant, and symbol handling.

	Author	:	Alex Stojan
*/

#ifndef _SYMBOLS_
#define _SYMBOLS_

#include <string>
#include <vector>
#include <map>
#include <stack>

#include <boost/lexical_cast.hpp>

#include "Globals.h"
#include "Utilities.h"

//*******************************************************************
//*******************************************************************
//*******************************************************************

namespace Compiler {

	class TypeAttr {
	public:
		typedef std::vector<int> DimVector;
		
		enum TypeSizeSpec {
			REGULAR,
			ARRAY_AS_POINTER		// array parameters are passed by reference, so they count as 4 bytes
		};

		TypeAttr():
			isPtr(false),
			refLevel(0),
			code(UNDEF_TYPE) {}
			
		TypeAttr(const TypeAttr& type):
			typeName(type.typeName),
			isPtr(type.isPointer()),
			refLevel(type.refLevel),
			arrayDimensions(type.arrayDimensions),
			size(type.size),
			code(type.code) {}
		
		explicit TypeAttr(std::string name, DimVector arrayDim = DimVector()):
			typeName(name),
			isPtr(false),
			refLevel(0),
			arrayDimensions(arrayDim) {

			calcSize(name);
			setTypeNameCode();
		}

		bool operator==(const TypeAttr& t) const {
			return typeName == t.typeName &&	
						isPtr == t.isPtr && 
						refLevel == t.refLevel &&
						arrayDimensions == t.arrayDimensions;
		}
		
		bool operator!=(const TypeAttr& t) const { return !operator==(t); }
		
		bool isAssignableTo(const TypeAttr& t) {
			return (*this == t) ||
						(isPrimitiveTypeAssignable(t.getTypeNameCode(), code) &&	
						(isArray() && t.isPtr && t.refLevel == 1 && isOneDimensionalArray()));
		}

		void clearArrayAttributes() { arrayDimensions.clear(); }

		// This is used for keeping track of type when indexing an array. For
		// example, for array {int[5][5] a}, 'a' is initially a two-dimensional
		// array; after writing a[2], the resulting type is a one-dimensional
		// array, and after writing a[2][3] the result is of type 'int'.
		//
		void reduceArray();

		bool isArray() const { return arrayDimensions.size() > 0; }
		bool isOneDimensionalArray() const { return arrayDimensions.size() == 1; }
		bool isCharString() const { return isOneDimensionalArray() && typeName == CHAR; }
		bool isIndexablePointer() const { return isPtr && refLevel == 1; }
		bool setTypeName(std::string, DimVector = DimVector());
		int getSize(SizeMetric = BYTE, TypeSizeSpec = REGULAR) const;
		TypeCode getTypeNameCode() const { return code; }
		void setAsPointer() { isPtr = true; }
		bool isPointer() const { return isPtr; }
		const DimVector& getArrayDim() const { return arrayDimensions; }
		
		// Since some instructions depend on the type and size of data (like
		// BLOAD, WLOAD for loading bytes and words), this method returns that
		// information about the type. The parameter specifies whether we need
		// a storage class of a pointer indexed as an array, in which case the 
		// storage class is the type of the pointer, not the pointer storage
		// class (usually the same as INT).
		//
		TypeCode getStorageClass(bool = false) const;
		
		bool isNumeric() const { return *this == TypeAttr(INT) || *this == TypeAttr(FLOAT); }
		bool isNumericOrChar() const { return *this == TypeAttr(INT) || *this == TypeAttr(FLOAT) || *this == TypeAttr(CHAR); }

		// returns the name of the base type; for 'int^' the base type is 'int'
		std::string getName() const { return typeName; }

		// The following two methods are used for adjusting the indirection level
		// of pointers; for example, if we have {int^^ p} then ^p would reduce
		// the resulting type to int^. Likewise, for {int t}, @t would increase
		// the indirection level by one, resulting in type int^.
		//
		bool reduceRefLevel();
		void increaseRefLevel();

		int getRefLevel() const { return refLevel; }
		
	private:
		void calcSize(std::string);
		bool setTypeNameCode();

		bool				isPtr;
		std::string		typeName;
		int				size;
		TypeCode			code;
		int				refLevel;		// 1 for T^, 2 for T^^, 3 for T^^^, etc.
		DimVector		arrayDimensions;
	};

	//****************************************************************

	struct Constant {
		Constant() {}
		
		explicit Constant(int n):
			token(boost::lexical_cast<std::string>(n)), 
			type(TypeAttr(INT)),
			intVal(n), 
			floatVal(0), 
			index(0) {}
		
		Constant(std::string s, TypeAttr t):
			token(s), 
			type(t),
			intVal(0), 
			floatVal(0), 
			index(0) {}

		Constant(int n, TypeAttr t):
			token(boost::lexical_cast<std::string>(n)), 
			type(t),
			intVal(n), 
			floatVal(0), 
			index(0) {}

		Constant(float f, TypeAttr t):
			token(boost::lexical_cast<std::string>(f)), 
			type(t),
			intVal(0), 
			floatVal(f), 
			index(0) {}

		bool operator==(const Constant& c) const {
			return token == c.token && type == c.type && intVal == c.intVal && floatVal == c.floatVal;
		}

		std::string		token;
		TypeAttr			type;
		int				intVal;		// also used for char and bool values
		float				floatVal;
		int				index;
	};

	//****************************************************************

	struct Symbol {
		typedef std::vector<Symbol*>	ParamVector;
		typedef std::vector<TypeAttr>	ForwardParamTypes;

		Symbol():
			isFunction(false),
			isLocal(false),
			paramsSpace(0),
			localSpace(0),
			isParameter(false),
			isFullFunctionDef(false),
			isForwardedFunction(false),
			index(0),
			localAddressCounter(0) {}

		Symbol(std::string sn, int id, TypeAttr t, ParamVector prm = ParamVector()):
			name(sn),
			symbolTableID(id),
			type(t), 
			isFunction(false),
			isLocal(false),
			params(prm),
			paramsSpace(0),
			localSpace(0),
			isParameter(false),
			isFullFunctionDef(false),
			isForwardedFunction(false),
			index(0),
			localAddressCounter(0) {}
			
		bool operator==(const Symbol& sym) { return name == sym.name; }
		bool operator!=(const Symbol& sym) { return !operator==(sym); }

		void setParams(ParamVector);

		// Name of the symbol as defined in the source code.
		std::string name;
		
		// The ID of the symbol table in which this symbol is stored. Since
		// scopes can be nested there could be more than one symbol with the
		// same name (hiding the outer one). This field is used by the assembly
		// generation function to obtain the index of the symbol stored in the
		// symbolic instruction object. For example, WLOAD "a" would require
		// looking up symbol "a" in the symbol table specified by this field of
		// the symbol object.
		int symbolTableID;

		// Type information for the symbol
		TypeAttr type;
		
		// Set to true if the symbol represents a function
		bool isFunction;
		
		// Set to True if the symbol is local
		bool isLocal;

		// Vector of parameter symbols (if the symbol is a function)
		ParamVector params;
		
		// Parameter types of forwarded functions
		ForwardParamTypes fwdParamTypes;
		
		// The amount of space occupied by parameters (used in function calls)
		int paramsSpace;
		
		// Total space (in words) occupied by local data (no constants)
		int localSpace;

		// Used for deciding whether to treat arrays as pointers
		bool isParameter;

		// This field is set to True after a function body has been processed
		// completely. It's used for deciding how to increment a stack pointer
		// for local data when calling a function (forward or normal call).
		bool isFullFunctionDef;
		
		// This field is set to True when the symbol represents a forwarded
		// function (declaration). Forwarded declarations are overwritten in the
		// symbol table by normal function definitions.
		bool isForwardedFunction;

		// Address of the symbol - in a case of function this field contains the
		// address of the first instruction
		int index;

		// This field is used to assign addresses to local symbols. Since a
		// function might contain nested blocks each symbol table for every block
		// will contain a pointer to the symbol representing the function. This
		// field is initialized to the return type size and is incremented for
		// each local symbol found in a symbol table that points to this symbol
		// object.
		int localAddressCounter;
	};

	//****************************************************************

	class SymbolTable {
	public:
		typedef std::map<std::string, Symbol>		SymbolMap;
		typedef std::map<std::string, Symbol*>		SymbolPtrMap;
		typedef std::vector<Symbol*>					SymbolSequence;
		typedef std::map<std::string, Constant>	ConstMap;

		SymbolTable();
		~SymbolTable();
		
		Symbol* addSymbol(std::string, TypeAttr, bool = false);
		void addConstant(Constant);
		Symbol* getSymbol(std::string);
		Symbol* getFunction(std::string);
		int getID() const { return id; }
		int getSpace() const { return space; }
		bool addFunction(Symbol*);
		const SymbolMap& getVarTable() const { return varTable; }
		const SymbolPtrMap& getFnTable() const { return fnTable; }
		const SymbolSequence& getVarSeq() const { return varSeq; }

		Symbol* function;		// function whose symbols this table holds

		// *** static members ***
		
		static Constant getConstant(std::string token) { return constTable[token]; }
		static ConstMap constTable;		// all constants are stored in the same map

	private:
		SymbolMap varTable;	// variables
		
		// This map contains function symbols only. Function symbols are stored
		// in a separate map from other symbols (variables) because of the way
		// the AST processes forward declarations and function definitions - when
		// a forward declaration is reached in AST the function symbol would 
		// normally be entered into the symbol table immediately, so it's not 
		// possible to verify that its signature matches the function definition
		// (or check for duplicate declarations, definitions, etc.) because the
		// parameters have not been reached yet. This would require a separate 
		// check at the end of the whole compilation process, when all the 
		// parameters are known.
		//
		SymbolPtrMap fnTable;

		// This vector holds pointers to symbols in unsorted order. It is used
		// when generating data headers for assembly files.
		//
		SymbolSequence varSeq;

		// Every symbol contains the ID of the symbol table where it's kept.
		//
		int id;
		
		// Total space (in words) adjusted every time a symbol or function is
		// added.
		//
		int space;

		// *** static members ***
		
		static int ID;
	};

	//****************************************************************

	// This class is used for handling entering and exiting scopes (like blocks
	// and functions). There is only one object of this class.
	//
	class Scope {
	public:
		typedef std::map<int, SymbolTable> SymbolTableMap;
		
		// This map holds all symbol tables created for a source file. It is used
		// when generating assembly headers to assign indexes to variables.
		//
		SymbolTableMap allSymbolTables;
		
		Scope::Scope(): space(0), topLevel(0) {};
		
		SymbolTable* enter(Symbol*);
		int exit();
		const std::stack<SymbolTable*>& getStack() const { return stStack; }
		SymbolTable* top() const { return stStack.top(); }
		SymbolTable* getTopLevel() const { return topLevel; }
		void resetSpace() { space = 0; }
		
	private:
		SymbolTable* topLevel;

		// This is used for implementing nested blocks, which also includes
		// function blocks because the outer scope of a function is the scope
		// global to the whole file (the global scope).
		//
		std::stack<SymbolTable*> stStack;

		// This field contains the amount of total space (in words) needed for
		// the current scope and all sub-scopes.
		//
		int space;
	};

	//****************************************************************

	// these two generate assembly file headers (constants and data)
	void generateConstantHeader(int*, std::vector<std::string>*, std::ofstream&);
	void generateDataHeader(const Scope::SymbolTableMap&, std::ofstream&, std::ofstream&);

	Symbol* findFunction(std::string, const Scope&);
	Symbol* findSymbol(std::string, int, Scope::SymbolTableMap&);
	Symbol* findSymbol(std::string, const std::stack<SymbolTable*>&);
	Constant getNumericConstant(std::string);

	// returns True if a forward declaration matches the function definition
	bool matchFwdToDef(const Symbol&, const Symbol&);
	
	// output functions (for info and debugging)

	std::ostream& operator<<(std::ostream&, const TypeAttr&);
	std::ostream& operator<<(std::ostream&, const Symbol&);
	std::ostream& operator<<(std::ostream&, SymbolTable&);
	std::ostream& operator<<(std::ostream&, const Constant&);
}

#endif
